@@ -1,5 +1,5 @@
 /**********************************************************************
-** $Id: qcanvas.cpp,v 1.2 2000/03/04 19:33:13 hadacek Exp $
+** $Id: qcanvas.cpp,v 1.4 2000/06/02 07:05:39 jones Exp $
 **
 ** Implementation of QCanvas and associated classes
 **
@@ -443,7 +443,8 @@ void QCanvas::init(int w, int h, int chunksze, int mxclusters)
 Create a QCanvas with no size.
 You will want to call resize(int,int) at some time after creation.
 */
-QCanvas::QCanvas()
+QCanvas::QCanvas( QObject* parent, const char* name )
+    : QObject( parent, name )
 {
     init(0,0);
 }
@@ -644,6 +645,26 @@ void QCanvas::retune(int chunksze, int mxclusters)
   Returns the size of the canvas, in pixels.
 */
 
+/*!
+  \fn bool QCanvas::onCanvas( int x, int y ) const
+  Returns whether the pixel position (\a x, \a y) is on the canvas.
+*/
+
+/*!
+  \fn bool QCanvas::onCanvas( const QPoint& p ) const
+  Returns whether the pixel position \a p is on the canvas.
+*/
+
+/*!
+  \fn bool QCanvas::validChunk( int x, int y ) const
+  Returns whether the chunk position (\a x, \a y) is on the canvas.
+*/
+
+/*!
+  \fn bool QCanvas::validChunk( const QPoint& p ) const
+  Returns whether the chunk position \a p is on the canvas.
+*/
+
 /*!
   \fn int QCanvas::chunkSize() const
   Returns the chunk size of the canvas as set at construction.
@@ -800,7 +821,7 @@ void QCanvas::update()
 
     for (QCanvasView* view=d->viewList.first(); view != 0; view=d->viewList.next()) {
 	QRect area(view->contentsX(),view->contentsY(),
-		view->contentsWidth(),view->contentsHeight());
+		view->visibleWidth(),view->visibleHeight());
 	if (area.width()>0 && area.height()>0) {
 	    clusterizer.add(area);
 	}
@@ -904,6 +925,8 @@ void QCanvas::drawArea(const QRect& inarea, QPainter* p, bool double_buffer)
 
     QCanvasItemList allvisible;
 
+    // Stores the region within area that need to be drawn. It is relative
+    // to area.topLeft()  (so as to keep within bounds of 16-bit XRegions)
     QRegion rgn;
 
     for (int x=lx; x<=mx; x++) {
@@ -916,7 +939,8 @@ void QCanvas::drawArea(const QRect& inarea, QPainter* p, bool double_buffer)
 	    if (!p) {
 		if ( chunk(x,y).takeChange() ) {
 		    // XXX should at least make bands
-		    rgn |= QRegion(x*chunksize,y*chunksize,chunksize,chunksize);
+		    rgn |= QRegion(x*chunksize-area.x(),y*chunksize-area.y(),
+				    chunksize,chunksize);
 		    allvisible += *chunk(x,y).listPtr();
 		}
 	    } else {
@@ -932,8 +956,14 @@ void QCanvas::drawArea(const QRect& inarea, QPainter* p, bool double_buffer)
 	int osh = area.height();
 	if ( osw > offscr.width() || osh > offscr.height() )
 	    offscr.resize(QMAX(osw,offscr.width()),QMAX(osh,offscr.height()));
+	else if ( offscr.width() == 0 || offscr.height() == 0 )
+	    offscr.resize( QMAX( offscr.width(), 1), QMAX( offscr.height(), 1 ) );
 	painter.begin(&offscr);
 	painter.translate(-area.x(),-area.y());
+	if ( p )
+	    painter.setClipRect(QRect(0,0,area.width(),area.height()));
+	else
+	    ;//painter.setClipRegion(rgn);
 	drawBackground(painter,area);
 	allvisible.drawUnique(painter);
 	drawForeground(painter,area);
@@ -947,10 +977,13 @@ void QCanvas::drawArea(const QRect& inarea, QPainter* p, bool double_buffer)
 	drawBackground(*p,area);
 	allvisible.drawUnique(*p);
 	drawForeground(*p,area);
+	return;
     }
 
     QPoint trtr; // keeps track of total translation of rgn
 
+    trtr -= area.topLeft();
+
     for (QCanvasView* view=d->viewList.first(); view; view=d->viewList.next()) {
 	QPainter painter(view->viewport());
 	QPoint tr = view->contentsToViewport(area.topLeft());
@@ -962,12 +995,13 @@ void QCanvas::drawArea(const QRect& inarea, QPainter* p, bool double_buffer)
 	    painter.setClipRegion(rgn);
 	    painter.drawPixmap(tr,offscr, QRect(QPoint(0,0),area.size()));
 	} else {
-	    painter.translate(rtr.x(),rtr.y());
+	    painter.translate(nrtr.x(),nrtr.y());
 	    rgn.translate(rtr.x(),rtr.y());
 	    painter.setClipRegion(rgn);
 	    drawBackground(painter,area);
 	    allvisible.drawUnique(painter);
 	    drawForeground(painter,area);
+	    painter.translate(-nrtr.x(),-nrtr.y());
 	}
     }
 }
@@ -985,7 +1019,7 @@ instead.
 */
 void QCanvas::setChangedChunk(int x, int y)
 {
-    if (x>=0 && x<chwidth && y>=0 && y<chheight) {
+    if (validChunk(x,y)) {
 	QCanvasChunk& ch=chunk(x,y);
 	ch.change();
     }
@@ -1017,7 +1051,7 @@ and SetChangedChunkContaining, this method marks the chunk as `dirty'.
 */
 void QCanvas::addItemToChunk(QCanvasItem* g, int x, int y)
 {
-    if (x>=0 && x<chwidth && y>=0 && y<chheight) {
+    if (validChunk(x,y)) {
 	chunk(x,y).add(g);
     }
 }
@@ -1030,7 +1064,7 @@ and SetChangedChunkContaining, this method marks the chunk as `dirty'.
 */
 void QCanvas::removeItemFromChunk(QCanvasItem* g, int x, int y)
 {
-    if (x>=0 && x<chwidth && y>=0 && y<chheight) {
+    if (validChunk(x,y)) {
 	chunk(x,y).remove(g);
     }
 }
@@ -1332,7 +1366,10 @@ Destructs the QCanvasItem.  It is removed from its canvas.
 */
 QCanvasItem::~QCanvasItem()
 {
-    if (cnv) cnv->removeItem(this);
+    if (cnv) {
+	cnv->removeItem(this);
+	cnv->removeAnimation(this);
+    }
     delete ext;
 }
 
@@ -1948,7 +1985,7 @@ QCanvasItemList QCanvasItem::collisions(bool exact) const
   Returns a list of items which intersect with the point \a p,
   sorted from shallowest to deepest.
 */
-QCanvasItemList QCanvas::collisions(QPoint p) const
+QCanvasItemList QCanvas::collisions(const QPoint& p) const
 {
     return collisions(QRect(p,QSize(1,1)));
 }
@@ -1957,7 +1994,7 @@ QCanvasItemList QCanvas::collisions(QPoint p) const
   Returns a list of items which intersect with the rectangle \a r,
   sorted from shallowest to deepest.
 */
-QCanvasItemList QCanvas::collisions(QRect r) const
+QCanvasItemList QCanvas::collisions(const QRect& r) const
 {
     QCanvasRectangle i(r,(QCanvas*)this);
     QCanvasItemList l = i.collisions(TRUE);
@@ -1975,7 +2012,7 @@ QCanvasItemList QCanvas::collisions(QRect r) const
   This is a utility function mainly used to implement the simpler
   QCanvasItem::collisions() function.
 */
-QCanvasItemList QCanvas::collisions(QPointArray chunks,
+QCanvasItemList QCanvas::collisions(const QPointArray& chunks,
 	    const QCanvasItem* item, bool exact) const
 {
     QPtrDict<void> seen;
@@ -1983,13 +2020,15 @@ QCanvasItemList QCanvas::collisions(QPointArray chunks,
     for (int i=0; i<(int)chunks.count(); i++) {
 	int x = chunks[i].x();
 	int y = chunks[i].y();
-	const QCanvasItemList* l = chunk(x,y).listPtr();
-	for (QCanvasItemList::ConstIterator it=l->begin(); it!=l->end(); ++it) {
-	    QCanvasItem *g=*it;
-	    if ( g != item ) {
-		if ( !seen.find(g) && (!exact || item->collidesWith(g)) ) {
-		    seen.replace(g,(void*)1);
-		    result.append(g);
+	if ( validChunk(x,y) ) {
+	    const QCanvasItemList* l = chunk(x,y).listPtr();
+	    for (QCanvasItemList::ConstIterator it=l->begin(); it!=l->end(); ++it) {
+		QCanvasItem *g=*it;
+		if ( g != item ) {
+		    if ( !seen.find(g) && (!exact || item->collidesWith(g)) ) {
+			seen.replace(g,(void*)1);
+			result.append(g);
+		    }
 		}
 	    }
 	}
@@ -2082,7 +2121,6 @@ QRect QCanvasItem::boundingRectAdvanced() const
   Constructs a QCanvasPixmap from an image file by loading it.
 */
 QCanvasPixmap::QCanvasPixmap(const QString& datafilename)
-: collision_mask(0)
 {
     QImage image(datafilename);
     init(image);
@@ -2091,7 +2129,6 @@ QCanvasPixmap::QCanvasPixmap(const QString& datafilename)
   Constructs a QCanvasPixmap from an image.
 */
 QCanvasPixmap::QCanvasPixmap(const QImage& image)
-: collision_mask(0)
 {
     init(image);
 }
@@ -2099,7 +2136,6 @@ QCanvasPixmap::QCanvasPixmap(const QImage& image)
   Constructs a QCanvasPixmap from a pixmap and an offset.
 */
 QCanvasPixmap::QCanvasPixmap(const QPixmap& pm, QPoint offset)
-: collision_mask(0)
 {
     init(pm,offset.x(),offset.y());
 }
@@ -2109,8 +2145,8 @@ void QCanvasPixmap::init(const QImage& image)
     convertFromImage(image);
     hotx = image.offset().x();
     hoty = image.offset().y();
-    if ( image.hasAlphaBuffer() )
-	collision_mask = new QImage(image.createAlphaMask());
+    collision_mask = image.hasAlphaBuffer() ?
+	new QImage(image.createAlphaMask()) : 0;
 }
 
 void QCanvasPixmap::init(const QPixmap& pixmap, int hx, int hy)
@@ -2118,8 +2154,8 @@ void QCanvasPixmap::init(const QPixmap& pixmap, int hx, int hy)
     (QPixmap&)*this = pixmap;
     hotx = hx;
     hoty = hy;
-    if ( pixmap.mask() )
-	collision_mask = new QImage(mask()->convertToImage());
+    collision_mask = pixmap.mask() ?
+	new QImage(mask()->convertToImage()) : 0;
 }
 
 /*!
@@ -2551,7 +2587,10 @@ QCanvasView::QCanvasView(QCanvas* canvas, QWidget* parent, const char* name, WFl
 {
     viewing = 0;
     setCanvas(canvas);
-    viewport()->setBackgroundColor(viewing->backgroundColor());
+
+    if ( viewing )
+	viewport()->setBackgroundColor(viewing->backgroundColor());
+
     connect(this,SIGNAL(contentsMoving(int,int)),this,SLOT(cMoving(int,int)));
 }
 
@@ -2585,6 +2624,8 @@ void QCanvasView::setCanvas(QCanvas* canvas)
 	viewing->addView(this);
     }
     updateContentsSize();
+    if ( viewing )
+	viewport()->setBackgroundColor(viewing->backgroundColor());
 }
 
 void QCanvasView::updateContentsSize()
@@ -2628,6 +2669,8 @@ void QCanvasView::drawContents(QPainter *p, int cx, int cy, int cw, int ch)
 */
 QSize QCanvasView::sizeHint() const
 {
+    if ( !canvas() )
+	return QScrollView::sizeHint();
     return canvas()->size()+QSize(frameWidth(),frameWidth())*2;
 }
 
